@@ -1,13 +1,13 @@
-import re
 import requests
 import logging
 import traceback
 import time
-from typing import List
+from typing import List, Dict, Tuple
 from datetime import datetime
 from bs4 import BeautifulSoup
-from cache import AdtCache
-from dartrig.annotations import memoize
+from cache import AdtCache, MemoryCache
+from cache.filecache import FileCache
+from dartrig.parser import extract_nodes, parse_dsaf, is_financial_company, DsafNode
 
 HEADERS = {
     "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36"}
@@ -20,14 +20,47 @@ URLS = {
 }
 logger = logging.getLogger("dartrig")
 
+def remove_strs_list(text, removes):
+    tmp = text
+    for remove in removes:
+        tmp = tmp.replace(remove, '')
+
+    return tmp
+
+def fun_splitter(s):
+    spl = s.split(" = ")
+    if len(spl) <= 1:
+        return None
+    try:
+        return int(spl[1])
+    except ValueError:
+        return None
+
+def get_node_text_contains(nodes, search_text):
+    for node in nodes:
+        node_txt = remove_strs_list(node.get("text"), [" "])
+        search_text = remove_strs_list(search_text, [" "])
+        if search_text in node_txt:
+            return node
+    return None
 
 class DartWeb:
-    def __init__(self):
+    def __init__(self, cache: AdtCache = None, file_cache: FileCache = None):
+        self.logger = logging.getLogger("dart_web")
         self.session = requests.Session()
         self.session.get(URLS["DART_BASE"], headers=HEADERS)
 
-    @memoize
-    def request_detail(self, rcp_no: str, dtd: str, ele_id: int = 0, offset: int = 0) -> str:
+        if cache is None:
+            self.logger.info("Cache is not provided. Use MemoryCache")
+            cache = MemoryCache()
+
+        self.cache: AdtCache = cache
+
+        self.file_cache = file_cache
+        if self.file_cache is not None:
+            self.logger.info(f"File cache is enabled. Cache directory: {file_cache.base_dir}")
+
+    def request_detail(self, rcp_no: str, dtd: str, ele_id: int = 0, offset: int = 0, length: int = 0) -> str:
         """
         :param rcp_no:
         :param dtd: html | dart3.xsd
@@ -35,47 +68,165 @@ class DartWeb:
         :param offset:
         :return:
         """
-        dcm_no = self.get_dcm_no_by_rcp_no(rcp_no)
-        url = f"{URLS['DART_VIEWER']}?rcpNo={rcp_no}&dcmNo={dcm_no}&eleId={ele_id}&offset={offset}&length=0&dtd={dtd}"
-        logger.info(f"request_detail url : {url}")
+        dcm_no = self._get_dcm_no_by_rcp_no(rcp_no)
+        return self.request_detail_with_dcm(rcp_no, dtd, dcm_no, ele_id, offset, length)
+
+    def request_detail_with_dcm(self, rcp_no, dtd, dcm_no, ele_id=0, offset=0, length=0):
+        dtd_split = dtd.split(".")[0]
+        file_cache_key = f"{rcp_no}_{dcm_no}_{dtd_split}_{ele_id}_{offset}"
+
+        url = f"{URLS['DART_VIEWER']}?rcpNo={rcp_no}&dcmNo={dcm_no}&eleId={ele_id}&offset={offset}&length={length}&dtd={dtd}"
+        self.logger.debug(f"request_detail_with_dcm url : {url}")
+        return self._try_file_cache_request(url, "viewer", file_cache_key, ext="html" if dtd.lower() == "html" else "xml")
+
+    def _try_file_cache_request(self, url, prefix, file_cache_key, ext="html"):
+        if self.file_cache is not None:
+            cached = self.file_cache.get_file(prefix=prefix, key=file_cache_key, ext=ext)
+            if cached is not None:
+                self.logger.debug(f"file cache hit for key {file_cache_key}")
+                return cached
+
         response = self.session.get(url, headers=HEADERS)
-        # response.encoding = "UTF-8"
-        return response.text
+        self.logger.debug(response.headers)
+        content_type = response.headers["Content-Type"]
+        if content_type is not None and "MS949" in content_type.upper():
+            self.logger.debug(f"response encoding is MS949. change to utf-8")
+            response.encoding = "ms949"
+            text = response.text
+        else:
+            text = response.text
 
-    @memoize
-    def request_detail_with_dcm(self, rcp_no, dtd, dcm_no, ele_id=0, offset=0):
-        url = f"{URLS['DART_VIEWER']}?rcpNo={rcp_no}&dcmNo={dcm_no}&eleId={ele_id}&offset={offset}&length=0&dtd={dtd}"
-        logger.info(f"request_detail_with_dcm url : {url}")
-        return self.session.get(url, headers=HEADERS).text
+        # self.logger.debug(f"_try_file_cache_request response text : {text}")
 
-    @memoize
-    def dsaf001_report(self, rcp_no):
-        logger.debug(f"dsaf001_report for rcp_no {rcp_no}")
-        return self.session.get(f"{URLS['DART_DSAF']}?rcpNo={rcp_no}", headers=HEADERS).text
+        if self.file_cache is not None:
+            self.file_cache.save_file(prefix=prefix, key=file_cache_key, ext=ext, data=text)
+            self.logger.debug(f"file cache set for key {file_cache_key}")
+        return text
 
-    @memoize
-    def get_dcm_no_by_rcp_no(self, rcp_no):
+    def get_dsaf_nodes(self, rcp_no) -> Tuple[str, List[DsafNode]]:
         try:
-            html_text = self.dsaf001_report(rcp_no)
-            numbers = re.findall("\d{7}", html_text)
-            dcm_no = numbers[2]
+            html_text = self._try_file_cache_request(f"{URLS['DART_DSAF']}?rcpNo={rcp_no}", "dsaf", rcp_no)
+            dcm_no, nodes = parse_dsaf(html_text)
+            return dcm_no, nodes
+        except Exception as ex:
+            logger.exception(ex)
+            raise ValueError(f"dcm number fetch failed for rcp_no : [{rcp_no}]")
+
+    def _get_dcm_no_by_rcp_no(self, rcp_no):
+        cache_key = f"dcm_no_{rcp_no}"
+        try:
+            if self.cache is not None:
+                cached = self.cache.get(cache_key)
+                if cached is not None:
+                    logger.debug(f"dsaf001_report cache hit for rcp_no {rcp_no}")
+                    return cached
+
+            dcm_no, _ = self.get_dsaf_nodes(rcp_no=rcp_no)
+
+            if self.cache is not None:
+                self.cache.set(cache_key, dcm_no)
+                logger.debug(f"dsaf001_report cache set for rcp_no {rcp_no}")
         except Exception as ex:
             logger.exception(ex)
             raise ValueError(f"dcm number fetch failed for rcp_no : [{rcp_no}]")
 
         return dcm_no
 
-    def get_document(self, rcp_no, dtd, ele_id=0, offset=0):
+
+    def get_dsaf_meta(self, company, rcp_no) -> Dict[str, any]:
+        """
+        :param rcp_no:
+        :param keywords: [연결재무, 재무에관한]
+        :return:
+        """
+        cache_key = f"dcm_meta_{rcp_no}"
+        try:
+            if self.cache is not None:
+                cached: Dict = self.cache.hget(cache_key)
+                if cached is not None and len(cached) > 0:
+                    logger.debug(f"dsaf001_meta cache hit for rcp_no {rcp_no} => {cached}")
+                    return cached
+
+            dcm_no, nodes = self.get_dsaf_nodes(rcp_no=rcp_no)
+
+            c_type = 'gb'
+            dcm_no, ele_id, offset, length = ['-1'] * 4
+            dtd = ""
+
+            if is_financial_company(company):
+                c_type = 'fin'
+
+                node_summ = get_node_text_contains(nodes, "1.요약재")
+                if node_summ:
+                    dcm_no = node_summ.get("dcmNo")
+                    ele_id = node_summ.get("eleId")
+                    offset = node_summ.get("offset")
+                    length = node_summ.get("length")
+                    dtd = node_summ.get("dtd")
+            else:
+                node = get_node_text_contains(nodes, "2.연결재무")
+                if node:
+                    node_desc = get_node_text_contains(nodes, "3.연결재무제표주석")
+                    if node_desc is None:
+                        c_type = 'yg'
+
+                        node_yg = get_node_text_contains(nodes, "2.연결재무제")
+                        if node_yg:
+                            dcm_no = node_yg.get("dcmNo")
+                            ele_id = node_yg.get("eleId")
+                            offset = node_yg.get("offset")
+                            length = node_yg.get("length")
+                            dtd = node_yg.get("dtd")
+                    else:
+                        node_abbr = get_node_text_contains(nodes, "1.요약재무")
+                        if node_abbr:
+                            node_small = get_node_text_contains(nodes, "소규모")
+                            node_hd = get_node_text_contains(nodes, "해당")
+                            node_conn = get_node_text_contains(nodes, "2.연결재무제표")
+
+                            if node_small or node_hd or node_conn:
+                                c_type = 'gb2'
+
+                                node_conn4 = get_node_text_contains(nodes, "4.재무제")
+                                if node_conn4:
+                                    dcm_no = node_conn4.get("dcmNo")
+                                    ele_id = node_conn4.get("eleId")
+                                    offset = node_conn4.get("offset")
+                                    length = node_conn4.get("length")
+                                    dtd = node_conn4.get("dtd")
+
+                if c_type == 'gb':
+                    node_ab = get_node_text_contains(nodes, "1.요약재")
+                    if node_ab:
+                        dcm_no = node_ab.get("dcmNo")
+                        ele_id = node_ab.get("eleId")
+                        offset = node_ab.get("offset")
+                        length = node_ab.get("length")
+                        dtd = node_ab.get("dtd")
+
+            meta = { "c_type": c_type, "dcm_no": dcm_no, "ele_id": ele_id, "offset": offset, "length": length, "dtd": dtd }
+
+            if self.cache is not None and ele_id != '-1':
+                self.cache.hset(cache_key, meta)
+                logger.debug(f"dsaf001_meta cache set for rcp_no {rcp_no}")
+
+            return meta
+        except Exception as ex:
+            logger.exception(ex)
+            raise ValueError(f"dsaf meta fetch failed for rcp_no : [{rcp_no}]")
+
+    def get_document(self, rcp_no, dtd, ele_id=0, offset=0, length=0):
         """
         문서 리턴
         :param rcp_no:
         :param dtd:
-        :param ele_id:
-        :param offset:
+        :param ele_id: 생략시 0
+        :param offset: 생략시 0
+        :param length: 생략시 0
         :return: (content_type, content)
         """
-        dcm_no = self.get_dcm_no_by_rcp_no(rcp_no)
-        url = f"{URLS['DART_VIEWER']}?rcpNo={rcp_no}&dcmNo={dcm_no}&eleId={ele_id}&offset={offset}&length=0&dtd={dtd}"
+        dcm_no = self._get_dcm_no_by_rcp_no(rcp_no)
+        url = f"{URLS['DART_VIEWER']}?rcpNo={rcp_no}&dcmNo={dcm_no}&eleId={ele_id}&offset={offset}&length={length}&dtd={dtd}"
         response = self.session.get(url, headers=HEADERS)
         content_type = response.headers.get("Content-Type")
         return content_type, response.content
@@ -181,7 +332,8 @@ class DartAPI:
                 keys = [x.get("rcept_no") for x in response_items]
                 diff = self.cache.differential(cache_key, keys)
                 logger.debug(f"diff : {diff}, cached keys : {len(self.cache.keys())}")
-                diff_ratio = float(len(diff)) / float(len(response_items)) * 100
+                diff_ratio = float(len(diff)) / float(len(response_items)) * 100 if len(response_items) > 0 else float(0)
+
                 if diff_ratio == float(0):
                     logger.info(f"diff ratio is {diff_ratio}% => break")
                     break
@@ -200,18 +352,21 @@ class DartAPI:
             else:
                 results.extend(response_items)
 
+                logger.info(f"pause {pause} secs for continue")
+                time.sleep(pause)
+
             if total_page == page_no:
                 logger.info(f"total page reached {total_page}")
                 break
 
-            if max_page >= page_no:
+            if max_page <= page_no:
                 logger.info(f"max page reached {max_page}")
                 break
 
         list_items = []
 
         for item in results:
-
+            logger.debug(f"item : {item}")
             # 각 항목별로 데이터 추출
             data = {
                 "company": item.get('corp_name', ''),
